use std::path::PathBuf;

use cargo_metadata::Metadata;
use clap::{Arg, ArgAction, Command};

mod subcommands {
    pub mod gen_cmake;
}

use subcommands::*;

// common options
const MANIFEST_PATH: &str = "manifest-path";
const CARGO_EXECUTABLE: &str = "cargo-executable";
const VERBOSE: &str = "verbose";
const LOCKED: &str = "locked";
const FROZEN: &str = "frozen";

pub struct GeneratorSharedArgs {
    pub manifest_path: PathBuf,
    pub cargo_executable: PathBuf,
    pub metadata: Metadata,
    pub verbose: bool,
}

fn main() -> Result<(), Box<dyn std::error::Error>> {
    let matches = Command::new("CMake Generator for Cargo")
        .version("0.1")
        .author("Andrew Gaspar <andrew.gaspar@outlook.com>")
        .about("Generates CMake files for Cargo projects")
        .arg(
            Arg::new(MANIFEST_PATH)
                .long("manifest-path")
                .value_name("Cargo.toml")
                .help("Specifies the target Cargo project")
                .required(true)
                .num_args(1),
        )
        .arg(
            Arg::new(CARGO_EXECUTABLE)
                .long("cargo")
                .value_name("EXECUTABLE")
                .required(true)
                .help("Path to the cargo executable to use"),
        )
        .arg(
            Arg::new(VERBOSE)
                .long("verbose")
                .action(ArgAction::SetTrue)
                .help("Request verbose output"),
        )
        .arg(
            Arg::new(LOCKED)
                .long("locked")
                .action(ArgAction::SetTrue)
                .help("Pass --locked to cargo invocations"),
        )
        .arg(
            Arg::new(FROZEN)
                .long("frozen")
                .action(ArgAction::SetTrue)
                .help("Pass --frozen to cargo invocations"),
        )
        .subcommand(gen_cmake::subcommand())
        .get_matches();

    let mut cmd = cargo_metadata::MetadataCommand::new();
    cmd.no_deps();
    if matches.get_flag(LOCKED) {
        cmd.other_options(["--locked".into()]);
    }
    if matches.get_flag(FROZEN) {
        cmd.other_options(["--frozen".into()]);
    }

    let manifest_path: &String = matches.get_one(MANIFEST_PATH).unwrap();
    let cargo_executable: &String = matches.get_one(CARGO_EXECUTABLE).unwrap();

    cmd.manifest_path(manifest_path);
    cmd.cargo_path(cargo_executable);

    let metadata = cmd.exec()?;

    let shared_args = GeneratorSharedArgs {
        manifest_path: manifest_path.into(),
        cargo_executable: cargo_executable.into(),
        metadata,
        verbose: matches.get_flag(VERBOSE),
    };

    match matches.subcommand() {
        Some((gen_cmake::GEN_CMAKE, matches)) => gen_cmake::invoke(&shared_args, matches)?,
        _ => unreachable!(),
    };

    // We should never reach this statement
    std::process::exit(1);
}
